"""
pysomneo - A Python library to interact with Philips Somneo devices.
"""

from .somneo import Somneo
from .const import (
    SOUND_SOURCE_ALARM,
    SOUND_SOURCE_DUSK,
    FM_PRESETS,
    SOURCES,
    DAYS,
    DAYS_TYPE,
    STATUS,
)

__all__ = [
    "Somneo",
    "SOUND_SOURCE_ALARM",
    "SOUND_SOURCE_DUSK",
    "FM_PRESETS",
    "SOURCES",
    "DAYS",
    "DAYS_TYPE",
    "STATUS",
]



"""
Utility functions for pysomneo.
"""

import calendar
from datetime import time, date, timedelta, datetime

from .const import DAYS, DAYS_TYPE, SOURCES


def days_int_to_list(days_int):
    """Convert integer to list of days."""
    if days_int == 0:
        return ["tomorrow"]
    return [v for k, v in DAYS.items() if k & days_int]


def days_list_to_int(days):
    """Convert list of days to integer."""
    return sum(k for k, v in DAYS.items() if v in days)


def days_int_to_type(days_int):
    """Convert integer to predefined days."""
    if days_int in DAYS_TYPE:
        return DAYS_TYPE[days_int]

    return "custom"


def alarms_to_dict(enabled_alarms, time_alarms):
    """Construct alarm data dictionary."""

    alarms = {}
    for alarm, enabled in enumerate(enabled_alarms["prfen"]):
        alarms[alarm] = {}
        alarms[alarm]["position"] = alarm + 1
        alarms[alarm]["name"] = "alarm" + str(alarm)
        alarms[alarm]["enabled"] = bool(enabled)
        alarms[alarm]["time"] = time(
            int(time_alarms["almhr"][alarm]), int(time_alarms["almmn"][alarm])
        )
        alarms[alarm]["days"] = days_int_to_list(int(time_alarms["daynm"][alarm]))
        alarms[alarm]["days_type"] = DAYS_TYPE.get(
            int(time_alarms["daynm"][alarm]), "custom"
        )
        alarms[alarm]["powerwake"] = bool(enabled_alarms["pwrsv"][3 * alarm])
        if bool(enabled_alarms["pwrsv"][3 * alarm]):
            alarms[alarm]["powerwake_delta"] = max(
                0,
                60 * int(enabled_alarms["pwrsv"][3 * alarm + 1])
                + int(enabled_alarms["pwrsv"][3 * alarm + 2])
                - 60 * int(time_alarms["almhr"][alarm])
                - int(time_alarms["almmn"][alarm]),
            )
        else:
            alarms[alarm]["powerwake_delta"] = 0

    return alarms


def sunset_to_dict(sunset_data, light_curves, sounds):
    """Construct sunset data dictionary."""
    data = {}
    data["is_on"] = bool(sunset_data["onoff"])
    data["duration"] = int(sunset_data["durat"])
    data["curve"] = list(light_curves.keys())[
        list(light_curves.values()).index(int(sunset_data["ctype"]))
    ]
    data["level"] = sunset_data["curve"]
    if sunset_data["snddv"] == "dus":
        data["sound"] = list(sounds.keys())[
            list(sounds.values()).index(int(sunset_data["sndch"]))
        ]
    elif sunset_data["snddv"] == "fmr":
        data["sound"] = "fm " + str(sunset_data["sndch"])
    elif sunset_data["snddv"] == "off":
        data["sound"] = "off"
    else:
        data["sound"] = sunset_data["sndch"]
    data["volume"] = sunset_data["sndlv"]

    return data


def player_to_dict(player, dusk_sound_themes):
    """Construct player data dictionary."""
    data = {}
    data["state"] = bool(player["onoff"])
    data["volume"] = (float(player["sdvol"]) - 1) / 24

    snddv = player.get("snddv")
    sndch = player.get("sndch")

    if player["snddv"] == "aux":
        data["source"] = "AUX"
    elif player["snddv"] == "fmr":
        data["source"] = "FM " + sndch
    elif snddv == "dus":
        theme_name = next(
            (k for k, v in dusk_sound_themes.items() if str(v) == str(sndch)),
            f"dusk:{sndch}",
        )
        data["source"] = theme_name.title()
    else:
        data["source"] = "Other"

    possible_sources = list(SOURCES.keys())
    if dusk_sound_themes:
        possible_sources += [name.title() for name in dusk_sound_themes.keys()]

    data["possible_sources"] = possible_sources

    return data


def get_next_alarm(alarms):
    """Get the next alarm that is set."""
    next_alarm = None
    new_next_alarm = None
    for alarm in alarms:
        if alarms[alarm]["enabled"] is True:
            # Get current time and day.
            now_time = datetime.now()
            now_day = date.today()

            # Get time and day of alarm
            alarm_time = alarms[alarm]["time"]
            alarm_days = alarms[alarm]["days"]

            # If alarm goes of tomorrow
            if alarm_days == ["tomorrow"]:
                alarm_time_full = datetime.combine(now_day, alarm_time)
                if alarm_time_full > now_time:
                    new_next_alarm = alarm_time_full
                else:
                    new_next_alarm = alarm_time_full + timedelta(days=1)
            # If days are specified
            else:
                # Find first following day that the alarm is set.
                for d in range(0, 7):
                    test_day = now_time.isoweekday() + d
                    if test_day > 7:
                        test_day -= 7
                    if calendar.day_abbr[test_day - 1].lower() in alarm_days:
                        alarm_time_full = datetime.combine(
                            now_day, alarm_time
                        ) + timedelta(days=d)
                        if alarm_time_full > now_time:
                            new_next_alarm = alarm_time_full
                            break

            if next_alarm:
                next_alarm = min(next_alarm, new_next_alarm)
            else:
                next_alarm = new_next_alarm

    if next_alarm:
        return next_alarm.astimezone()
    return None



"""
Main Somneo class, represents the state of the Somneo wake-up light
and provides methods to interact with it.
"""

import time
import logging
import datetime
import uuid

from .api import SomneoClient
from .const import DAYS_TYPE, SOUND_SOURCE_ALARM, STATUS
from .util import (
    alarms_to_dict,
    get_next_alarm,
    days_list_to_int,
    sunset_to_dict,
    player_to_dict,
)

_LOGGER = logging.getLogger(__name__)


class Somneo(object):
    """
    Class represents the Somneo wake-up light.
    """

    def __init__(self, host=None, use_session=True, fast_interval=5, slow_interval=60):
        """Initialize."""
        self._host = host
        self._client = SomneoClient(host=host, use_session=use_session)
        # The fastest refresh interval is only relevant for sensor data and is 5 seconds by default
        self.fast_interval = fast_interval
        # The slowest refresh interval is relevant for all other data and is 15 minutes by default
        self.slow_interval = slow_interval

        self._last_sensor_fetch = 0
        self._last_slow_fetch = 0

        self.data = {}

        self.alarm_status = None
        self.light_data = None
        self.sensor_data = None
        self.sunset_data = None
        self.enabled_alarms = None
        self.time_alarms = None
        self.snoozetime = None
        self.player = None
        self._wake_light_themes = {}
        self._dusk_light_themes = {}
        self._wake_sound_themes = {}
        self._dusk_sound_themes = {}

    @property
    def wake_light_themes(self):
        """Get valid light curves for this light."""
        if len(self._wake_light_themes) == 0:
            self._fetch_themes()
            _LOGGER.debug(self._wake_light_themes)
        return self._wake_light_themes

    @property
    def dusk_light_themes(self):
        """Get valid dusk curves for this light."""
        if len(self._dusk_light_themes) == 0:
            self._fetch_themes()
            _LOGGER.debug(self._dusk_light_themes)
        return self._dusk_light_themes

    @property
    def wake_sound_themes(self):
        """Get valid wake-up sounds for this light."""
        if len(self._wake_sound_themes) == 0:
            self._fetch_themes()
            _LOGGER.debug(self._wake_sound_themes)
        return self._wake_sound_themes

    @property
    def dusk_sound_themes(self):
        """Get valid winddown sounds for this light."""
        if len(self._dusk_sound_themes) == 0:
            self._fetch_themes()
            _LOGGER.debug(self._dusk_sound_themes)
        return self._dusk_sound_themes

    def _fetch_themes(self):
        """Get themes."""
        themes = self._client.get_themes()
        self._wake_light_themes = themes["wake_light"]
        self._dusk_light_themes = themes["dusk_light"]
        self._wake_sound_themes = themes["wake_sound"]
        self._dusk_sound_themes = themes["dusk_sound"]
        self._fetch_sunset_data()

    def get_device_info(self):
        """Get device information via SomneoClient, fallback to defaults if unavailable."""
        # Default values if XML fetch fails
        device_info = {
            "manufacturer": "Royal Philips Electronics",
            "model": "Wake-up Light",
            "modelnumber": "Unknown",
            "serial": str(uuid.uuid1()),
        }

        # Use the client to get the XML as an ElementTree root
        root = self._client.get_description_xml()
        if root is not None:
            try:
                # Map XML elements to device_info
                device_info["manufacturer"] = root[1][2].text
                device_info["model"] = root[1][3].text
                device_info["modelnumber"] = root[1][4].text
                device_info["serial"] = root[1][6].text
            except (IndexError, AttributeError) as e:
                _LOGGER.warning(
                    "Failed to parse XML elements, using default device info: %s", e
                )

        _LOGGER.debug("Device info: %s", device_info)
        return device_info

    def fetch_data(self, force_slow_refresh=False):
        """Retrieve information from Somneo"""
        _LOGGER.debug("Calling Somneo.fetch_data()")
        now = time.time()

        # Sensor data is usefull to fetch more often
        if now - self._last_sensor_fetch >= self.fast_interval:
            self._fetch_sensor_data()
            self._fetch_alarm_status()
            self._last_sensor_fetch = now

        if now - self._last_slow_fetch >= self.slow_interval or force_slow_refresh:
            self._fetch_light_data()
            self._fetch_sunset_data()
            self._fetch_alarm_data()
            self._fetch_snooze_time()
            self._fetch_player_data()
            self._last_slow_fetch = now

        return self.data

    def _update_sensor_data(self, sensor_data):
        """Update sensor data in data object"""
        self.data["temperature"] = sensor_data["temperature"]
        self.data["humidity"] = sensor_data["humidity"]
        self.data["luminance"] = sensor_data["luminance"]
        self.data["noise"] = sensor_data["noise"]

    def _fetch_sensor_data(self):
        """Fetch only the sensor data from Somneo"""
        sensor_data = self._client.get_sensor_data()
        _LOGGER.debug("Fetched sensor data: %s", sensor_data)
        self._update_sensor_data(sensor_data)

    def _update_light_data(self):
        """Update light data in data object"""
        self.data["light_is_on"] = bool(self.light_data["onoff"])
        self.data["light_brightness"] = int(int(self.light_data["ltlvl"]) / 25 * 255)
        self.data["nightlight_is_on"] = bool(self.light_data["ngtlt"])

    def _fetch_light_data(self):
        """Fetch only the light data from Somneo"""
        self.light_data = self._client.get_light_data()
        _LOGGER.debug("Fetched light data: %s", self.light_data)
        self._update_light_data()

    def _update_alarm_status(self):
        """Update alarm status in data object"""
        self.data["somneo_status"] = STATUS.get(self.alarm_status["wusts"], "unknown")
        self.data["display_always_on"] = bool(self.alarm_status["dspon"])
        self.data["display_brightness"] = int(self.alarm_status["brght"])

    def _fetch_alarm_status(self):
        """Fetch only the alarm status from Somneo"""
        self.alarm_status = self._client.get_alarm_status()
        _LOGGER.debug("Fetched alarm status: %s", self.alarm_status)
        self._update_alarm_status()

    def _update_sunset_data(self):
        """Update sunset data in data object"""
        self.data["sunset"] = sunset_to_dict(
            self.sunset_data, self.dusk_light_themes, self.dusk_sound_themes
        )

    def _fetch_sunset_data(self):
        """Fetch only the sunset data from Somneo"""
        self.sunset_data = self._client.get_sunset_data()
        _LOGGER.debug("Fetched sunset data: %s", self.sunset_data)
        self._update_sunset_data()

    def _update_alarm_data(self):
        """Update alarm data in data object"""
        self.data["alarms"] = alarms_to_dict(self.enabled_alarms, self.time_alarms)
        self.data["next_alarm"] = get_next_alarm(self.data["alarms"])

    def _fetch_alarm_data(self):
        """Fetch only the alarm data from Somneo"""
        self.enabled_alarms = self._client.get_enabled_alarms()
        self.time_alarms = self._client.get_time_alarms()
        _LOGGER.debug("Fetched enabled alarms: %s", self.enabled_alarms)
        _LOGGER.debug("Fetched time alarms: %s", self.time_alarms)
        self._update_alarm_data()

    def _update_snooze_time(self):
        """Update snooze time in data object"""
        self.data["snooze_time"] = self.snoozetime["snztm"]

    def _fetch_snooze_time(self):
        """Fetch only the snooze time from Somneo"""
        self.snoozetime = self._client.get_snooze_time()
        _LOGGER.debug("Fetched snooze time: %s", self.snoozetime)
        self._update_snooze_time()

    def _update_player_data(self):
        """Update player data in data object"""
        self.data["player"] = player_to_dict(self.player, self.dusk_sound_themes)

    def _fetch_player_data(self):
        """Fetch only the player data from Somneo"""
        self.player = self._client.get_player_status()
        _LOGGER.debug("Fetched player status: %s", self.player)
        self._update_player_data()

    def toggle_light(self, state, brightness=None):
        """Toggle the light on or off"""
        if not self.light_data:
            self._fetch_light_data()

        payload = dict(self.light_data)
        payload["onoff"] = state
        payload["ngtlt"] = False
        if brightness:
            payload["ltlvl"] = int(brightness / 255 * 25)

        _LOGGER.debug("PUT toggle_light payload=%s", payload)
        response = self._client.modify_light(payload=payload)
        _LOGGER.debug("PUT toggle_light response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_light_data()
        self._fetch_sensor_data()

    def toggle_night_light(self, state):
        """Toggle the night light on or off"""
        if not self.light_data:
            self._fetch_light_data()

        payload = dict(self.light_data)
        payload["onoff"] = False
        payload["ngtlt"] = state

        _LOGGER.debug("PUT toggle_night_light payload=%s", payload)
        response = self._client.modify_light(payload=payload)
        _LOGGER.debug("PUT toggle_night_light response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_light_data()
        self._fetch_sensor_data()

    def dismiss_alarm(self):
        """Dismiss a running alarm."""
        payload = {"disms": True}
        _LOGGER.debug("PUT dismiss_alarm payload=%s", payload)
        response = self._client.modify_running_alarm(payload=payload)
        _LOGGER.debug("PUT dismiss_alarm response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def snooze_alarm(self):
        """Snooze a running alarm."""
        payload = {"tapsz": True}
        _LOGGER.debug("PUT snooze_alarm payload=%s", payload)
        response = self._client.modify_running_alarm(payload=payload)
        _LOGGER.debug("PUT snooze_alarm response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()
        self._fetch_snooze_time()

    def get_alarm_details(self, alarm):
        """Get the alarm settings."""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        alarm_pos = self.data["alarms"][alarm]["position"]
        payload = {"prfnr": alarm_pos}
        _LOGGER.debug("PUT get_alarm_details payload=%s", payload)
        response = self._client.modify_alarm_details(payload=payload)
        _LOGGER.debug("PUT get_alarm_details response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()
        return response

    def toggle_alarm(self, alarm, status):
        """Toggle the alarm on or off"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        payload = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "prfvs": True,
            "prfen": status,
        }
        _LOGGER.debug("PUT toggle_alarm payload=%s", payload)
        response = self._client.modify_alarm_wake_up_configuration(payload=payload)
        _LOGGER.debug("PUT toggle_alarm response=%s", response)

        self.data["alarms"][alarm]["enabled"] = status
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def set_alarm(self, alarm, v_time=None, days=None):
        """Set the time and day of an alarm."""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        alarm_settings = {"prfnr": self.data["alarms"][alarm]["position"]}
        if v_time is not None:
            alarm_settings["almhr"] = v_time.hour
            alarm_settings["almmn"] = v_time.minute
            self.data["alarms"][alarm]["time"] = v_time
        if days is not None:
            if isinstance(days, list):
                days_int = days_list_to_int(days)
            elif days in DAYS_TYPE.values():
                days_int = next(k for k, v in DAYS_TYPE.items() if v == days)
            else:
                days_int = int(self.time_alarms["daynm"][alarm])
            alarm_settings["daynm"] = days_int
            self.data["alarms"][alarm]["days"] = days
            self.data["alarms"][alarm]["days_type"] = DAYS_TYPE.get(days_int, "custom")

        if self.data["alarms"][alarm]["powerwake"]:
            alarm_dt = datetime.datetime.strptime(
                self.data["alarms"][alarm]["time"].isoformat(), "%H:%M:%S"
            )
            pw_dt = alarm_dt + datetime.timedelta(
                minutes=self.data["alarms"][alarm]["powerwake_delta"]
            )
            alarm_settings["pszhr"] = pw_dt.hour
            alarm_settings["pszmn"] = pw_dt.minute

        _LOGGER.debug("PUT set_alarm payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT set_alarm response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def set_alarm_light(self, alarm, curve="sunny day", level=20, duration=30):
        """Adjust the lightcurve of the wake-up light"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()
        if not self.wake_light_themes:
            self._fetch_themes()

        alarm_settings = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "ctype": self.wake_light_themes[curve],
            "curve": level,
            "durat": duration,
        }

        _LOGGER.debug("PUT set_alarm_light payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT set_alarm_light response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()
        self._fetch_light_data()
        self._fetch_sensor_data()

    def set_alarm_sound(
        self, alarm, source="wake-up", channel="forest birds", level=12
    ):
        """Adjust the alarm sound of the wake-up light"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()
        if not self.wake_sound_themes:
            self._fetch_themes()

        alarm_settings = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "snddv": SOUND_SOURCE_ALARM[source],
            "sndch": (
                self.wake_sound_themes[channel]
                if source == "wake-up"
                else (" " if source == "off" else channel)
            ),
            "sndlv": level,
        }

        _LOGGER.debug("PUT set_alarm_sound payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT set_alarm_sound response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()
        self._fetch_player_data()
        self._fetch_sensor_data()

    def set_alarm_powerwake(self, alarm, onoff=False, delta=0):
        """Set power wake"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        alarm_datetime = datetime.datetime.strptime(
            self.data["alarms"][alarm]["time"].isoformat(), "%H:%M:%S"
        )
        powerwake_datetime = alarm_datetime + datetime.timedelta(minutes=delta)

        alarm_settings = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "pwrsz": 1 if onoff else 0,
            "pszhr": powerwake_datetime.hour if onoff else 0,
            "pszmn": powerwake_datetime.minute if onoff else 0,
        }

        self.data["alarms"][alarm]["powerwake"] = onoff
        self.data["alarms"][alarm]["powerwake_delta"] = delta if onoff else 0

        _LOGGER.debug("PUT set_alarm_powerwake payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT set_alarm_powerwake response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def set_snooze_time(self, snooze_time=9):
        """Adjust the snooze time (minutes) of all alarms"""
        payload = {"snztm": snooze_time}
        _LOGGER.debug("PUT set_snooze_time payload=%s", payload)
        response = self._client.modify_alarm_details(payload=payload)
        _LOGGER.debug("PUT set_snooze_time response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_snooze_time()

    def add_alarm(self, alarm):
        """Add alarm to the list"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        alarm_settings = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "prfvs": True,
        }

        _LOGGER.debug("PUT add_alarm payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT add_alarm response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def remove_alarm(self, alarm):
        """Remove alarm from the list"""
        if not self.enabled_alarms or not self.time_alarms:
            self._fetch_alarm_data()

        alarm_settings = {
            "prfnr": self.data["alarms"][alarm]["position"],
            "prfen": False,
            "prfvs": False,
            "almhr": 7,
            "almmn": 30,
            "pwrsz": 0,
            "pszhr": 0,
            "pszmn": 0,
            "ctype": 0,
            "curve": 20,
            "durat": 30,
            "daynm": 254,
            "snddv": "wus",
            "sndch": "1",
            "sndlv": 12,
        }

        _LOGGER.debug("PUT remove_alarm payload=%s", alarm_settings)
        response = self._client.modify_alarm_wake_up_configuration(
            payload=alarm_settings
        )
        _LOGGER.debug("PUT remove_alarm response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_data()

    def toggle_sunset(self, status):
        """Toggle the sunset feature on or off"""
        if not self.sunset_data:
            self._fetch_sunset_data()

        payload = {"onoff": status}
        _LOGGER.debug("PUT toggle_sunset payload=%s", payload)
        response = self._client.modify_sunset(payload=payload)
        _LOGGER.debug("PUT toggle_sunset response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_sunset_data()
        self._fetch_player_data()
        self._fetch_alarm_status()
        self._fetch_sensor_data()

    def set_sunset(
        self, curve=None, level=None, duration=None, sound=None, volume=None
    ):
        """Adjust the sunset settings"""
        if (
            not self.dusk_light_themes
            or not self.dusk_sound_themes
            or not self.sunset_data
        ):
            self._fetch_themes()

        sunset_settings = dict(self.sunset_data)

        if duration:
            sunset_settings["durat"] = duration
        if curve:
            sunset_settings["ctype"] = self.dusk_light_themes[curve.lower()]
        if level:
            sunset_settings["curve"] = level
        if sound:
            if sound == "off":
                sunset_settings["snddv"] = "off"
            elif sound.upper().startswith("FM"):
                sunset_settings["snddv"] = "fmr"
                sunset_settings["sndch"] = sound[3:]
            elif sound.lower() in self.dusk_sound_themes:
                sunset_settings["snddv"] = "dus"
                sunset_settings["sndch"] = self.dusk_sound_themes[sound.lower()]
            else:
                _LOGGER.error("Invalid sound specified: %s", sound)
                raise ValueError(f"Unsupported sunset sound: {sound}")
        if volume:
            sunset_settings["sndlv"] = volume

        if bool(sunset_settings["onoff"]):
            _LOGGER.debug(
                "Sunset is already on, to modify it " \
                "we need to turn it off first otherwise changes are ignored"
            )
            self.toggle_sunset(False)
            time.sleep(1)  # short delay to allow the device to process

        _LOGGER.debug("PUT set_sunset payload=%s", sunset_settings)
        response = self._client.modify_sunset(payload=sunset_settings)
        _LOGGER.debug("PUT set_sunset response=%s", response)
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        time.sleep(0.1)  # Short delay to allow the device to process
        self._fetch_sunset_data()
        self._fetch_player_data()
        self._fetch_alarm_status()
        self._fetch_sensor_data()

    def toggle_player(self, state: bool):
        """Toggle the audio player"""
        if not self.player:
            self._fetch_player_data()

        data = dict(self.player)
        data["onoff"] = state

        _LOGGER.debug("PUT toggle_player payload=%s", data)
        response = self._client.modify_player(payload=data)
        _LOGGER.debug("PUT toggle_player response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_player_data()
        # It might also affect the sunset state
        self._fetch_sunset_data()
        self._fetch_alarm_status()
        self._fetch_sensor_data()

    def set_player_volume(self, volume: float):
        """Set the volume of the player (0..1)"""
        volume = min(max(volume, 0), 1)

        payload = {"sdvol": int(volume * 24 + 1)}
        _LOGGER.debug("PUT set_player_volume payload=%s", payload)
        response = self._client.modify_player(payload=payload)
        _LOGGER.debug("PUT set_player_volume response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_player_data()
        self._fetch_sensor_data()

    def set_player_source(self, source: str):
        """
        Set the source of the player, either 'aux'
        or preset 1..5 or one of the dusk sound theme
        """
        if not self.player:
            self._fetch_player_data()

        previous_state = self.player["onoff"]
        previous_sndch = self.player["sndch"]
        sunset_settings = dict(self.sunset_data)

        snddv = None
        sndch = None

        if source.upper() == "AUX":
            snddv = "aux"
            sndch = "1"
        elif source.upper().startswith("FM "):
            snddv = "fmr"
            sndch = source.split(" ")[1]
        elif source.lower() in self.dusk_sound_themes:
            snddv = "dus"
            sndch = self.dusk_sound_themes[source.lower()]
        else:
            _LOGGER.error("Invalid source specified: %s", source)
            raise ValueError(f"Unsupported player source: {source}")

        payload = {
            "snddv": snddv,
            "sndch": sndch,
            "sndss": 0,
            "onoff": previous_state,
            "tempy": False,
        }

        _LOGGER.debug("PUT set_player_source payload=%s", payload)
        response = self._client.modify_player(payload=payload)
        _LOGGER.debug("PUT set_player_source response=%s", response)

        if bool(sunset_settings["onoff"]) and previous_sndch != sndch:
            _LOGGER.debug(
                "Sunset is already on and we modified the sound, " \
                "to apply these we need to modify sunset endpoint"
            )
            self.set_sunset(sound=source)

        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_player_data()
        # It might also affect the sunset state
        self._fetch_sunset_data()
        self._fetch_alarm_status()
        self._fetch_sensor_data()

    def set_display(self, state=None, brightness=None):
        """Adjust the display"""
        if not self.alarm_status:
            self._fetch_alarm_status()

        payload = {
            "dspon": state if state is not None else self.data["display_always_on"],
            "brght": (
                brightness
                if brightness is not None
                else self.data["display_brightness"]
            ),
        }

        _LOGGER.debug("PUT set_display payload=%s", payload)
        response = self._client.modify_alarm_status(payload=payload)
        _LOGGER.debug("PUT set_display response=%s", response)
        time.sleep(0.1)  # Short delay to allow the device to process
        # The response of the put command is incomplete, so sent a new request
        # before updating internal state
        self._fetch_alarm_status()
        self._fetch_sensor_data()



"""
Constants for the pysomneo library.
"""

SOUND_SOURCE_ALARM = {"wake-up": "wus", "radio": "fmr", "off": "off"}
SOUND_SOURCE_DUSK = {"radio": "fmr", "off": "off", "dusk": "dus"}

FM_PRESETS = ["fm 1", "fm 2", "fm 3", "fm 4", "fm 5"]

SOURCES = {
    "AUX": "aux",
    "FM 1": 1,
    "FM 2": 2,
    "FM 3": 3,
    "FM 4": 4,
    "FM 5": 5,
    "Other": "other",
}

DAYS = {
    0: "tomorrow",
    2: "mon",
    4: "tue",
    8: "wed",
    16: "thu",
    32: "fri",
    64: "sat",
    128: "sun",
}
DAYS_TYPE = {0: "tomorrow", 62: "workdays", 192: "weekend", 254: "daily"}

STATUS = {
    1: "off",
    2: "sunset",
    257: "light-on",
    776: "sunset",
    777: "sunset",
    2321: "snooze",
    2309: "wake-up",
    2817: "on",
}



"""
Philips Somneo API client using requests with connection pooling, retries, and session management.
"""

import time
import logging
import xml.etree.ElementTree as ET
from typing import Any
from urllib.parse import urljoin
import urllib3
from urllib3.util.retry import Retry
from urllib3.exceptions import NewConnectionError

from requests import Session, request
from requests.adapters import HTTPAdapter
from requests.exceptions import (
    ConnectTimeout,
    ReadTimeout,
    RequestException,
    Timeout,
    ConnectionError as RequestsConnectionError,
)

_LOGGER = logging.getLogger(__name__)


class SomneoInvalidURLError(RequestException):
    """Raised when the Somneo device responds with 422 Invalid URL."""


class SomneoSession(Session):
    """
    requests.Session subclass that:
      - supports a base_url,
      - mounts a pool adapter for connection reuse,
      - will reset its internal session/pool once on ConnectionError to recover stale sockets,
      - exposes a short retry/backoff strategy for transient errors.
    """

    def __init__(
        self,
        base_url: str | None = None,
        use_session: bool = True,
        connect_timeout: float = 2.0,
        read_timeout: float = 8.0,
        timeout: tuple[float, float] | None = None,
        pool_connections: int = 1,
        pool_maxsize: int = 1
    ):
        super().__init__()
        self.base_url = base_url
        self._use_session = use_session
        self._pool_connections = pool_connections
        self._pool_maxsize = pool_maxsize
        self._timeout = (connect_timeout, read_timeout) if timeout is None else timeout

        self._mount_adapter()

    def _mount_adapter(self):
        adapter = HTTPAdapter(
            pool_connections=self._pool_connections,
            pool_maxsize=self._pool_maxsize,
            max_retries=Retry(total=0),
            pool_block=False,
        )
        # (re)mount adapters for both http and https
        self.mount("http://", adapter)
        self.mount("https://", adapter)

    def _reset_session_pool(self):
        """Close the session and reinitialize internals and adapters."""
        try:
            # close underlying connections (releases pools)
            self.close()
        except Exception as exc:
            _LOGGER.debug("Error while closing session: %s", exc)

        # Re-init Session internals
        super().__init__()
        # Re-mount adapters after re-init
        self._mount_adapter()

    def _get_sleep_time(self, weight: float, attempt: int) -> float:
        """Calculate exponential backoff sleep time."""
        return min(weight * (2 ** (attempt - 1)), 10)

    def _classify_error(self, e):
        """
        Classify exceptions for logging, pool reset, and backoff.
        Returns: (err_type: str, reset_pool: bool, weight: float)
        """
        if isinstance(e, ConnectTimeout):
            return "ConnectTimeout", True, 1.5
        elif isinstance(e, ReadTimeout):
            return "ReadTimeout", False, 2.0
        elif isinstance(e, RequestsConnectionError):
            if isinstance(
                getattr(e, "__cause__", None), NewConnectionError
            ) or "NewConnectionError" in str(e):
                return "NewConnectionError", True, 1.5
            return "ConnectionError", True, 1.0
        elif isinstance(e, Timeout):
            return "Timeout", False, 0.5
        else:  # fallback for other RequestExceptions
            return "RequestException", False, 0.75

    def request(self, method, url, **kwargs):
        if self.base_url:
            full_url = urljoin(self.base_url, url)
        else:
            full_url = url

        if "timeout" not in kwargs:
            kwargs["timeout"] = self._timeout

        max_attempts = 3
        last_exc = None
        has_reset_pool = False

        for attempt in range(1, max_attempts + 1):
            try:
                if self._use_session:
                    resp = super().request(method, full_url, **kwargs)
                else:
                    resp = request(method, full_url, timeout=self._timeout, **kwargs)

                if resp.status_code == 422:
                    raise SomneoInvalidURLError(
                        f"Invalid URL: {full_url}", response=resp
                    )
                return resp

            except (
                ConnectTimeout,
                ReadTimeout,
                ConnectionError,
                Timeout,
                RequestException,
            ) as e:
                err_type, reset_pool, weight = self._classify_error(e)

                _LOGGER.debug(
                    "%s (attempt %d/%d) when calling %s: %s",
                    err_type,
                    attempt,
                    max_attempts,
                    full_url,
                    e,
                )
                last_exc = e

                # Reset pool if necessary
                if (
                    reset_pool
                    and attempt <= max_attempts
                    and self._use_session
                    and not has_reset_pool
                ):
                    _LOGGER.info(
                        "Resetting session pool (attempt %d) for %s due to %s",
                        attempt,
                        full_url,
                        err_type,
                    )
                    try:
                        self._reset_session_pool()
                        has_reset_pool = True
                        # longer backoff after reset to allow new connection to be established
                        weight = 8.0
                    except (OSError, RuntimeError) as exc:
                        _LOGGER.debug("Session reset failed: %s", exc)

                # Backoff
                sleep = self._get_sleep_time(weight, attempt)

                if attempt < max_attempts:
                    _LOGGER.debug(
                        "Sleeping %.2fs before retrying %s (attempt %d/%d)",
                        sleep,
                        full_url,
                        attempt,
                        max_attempts,
                    )
                    time.sleep(sleep)

        _LOGGER.info("All %d attempts failed for %s", max_attempts, full_url)
        if last_exc is not None:
            raise last_exc
        raise RequestException("Unknown error in SomneoSession.request")


class SomneoClient:
    """High-level client for interacting with the Philips Somneo API."""

    def __init__(self, host: str, use_session: bool = True):
        urllib3.disable_warnings()
        self.host = host
        self.timeout = (2.0, 8.0)  # (connect, read) timeouts in seconds
        self.session = SomneoSession(
            base_url=f"https://{host}/di/v1/products/1/",
            use_session=use_session,
            timeout=self.timeout,
        )

    def _internal_call(
        self,
        method: str,
        path: str,
        headers: dict[str, str] | None = None,
        payload: dict[str, Any] | None = None,
    ):
        """Internal call to the device reusing SomneoSession"""
        args: dict[str, Any] = {}
        if payload:
            args["json"] = payload
        if headers:
            args["headers"] = headers

        r = None
        try:
            r = self.session.request(
                method, path, verify=False, timeout=self.timeout, **args
            )
            r.raise_for_status()
            return r.json()
        finally:
            if r is not None:
                r.close()

    def _get(self, path: str):
        """Perform a GET request."""
        return self._internal_call("GET", path)

    def put(self, path: str, payload: dict[str, Any]) -> Any:
        """Perform a PUT request with JSON payload."""
        return self._internal_call("PUT", path, payload=payload)

    def get_description_xml(self):
        """
        Fetch the device description XML from the Somneo device.
        Tries HTTPS first, then HTTP as fallback.
        Returns raw XML content.
        """
        urls = [
            f"https://{self.host}/upnp/description.xml",
            f"http://{self.host}/upnp/description.xml",
        ]

        last_exc = None
        for url in urls:
            response = None
            try:
                response = self.session.request(
                    "GET", url, verify=False, timeout=self.timeout
                )
                response.raise_for_status()
                # Try parsing immediately to ensure it's valid XML
                root = ET.fromstring(response.content)
                return root

            except RequestException as e:
                _LOGGER.debug("Connection failed for %s: %s", url, e)
                last_exc = e
            except ET.ParseError as e:
                _LOGGER.debug("XML parsing failed for %s: %s", url, e)
                last_exc = e
            finally:
                if response is not None:
                    response.close()

        if last_exc is not None:
            raise last_exc

        # Return None if all attempts failed
        return None

    def get_themes(self) -> dict[str, dict[str, int]]:
        """Get available light and sound themes as a dictionary."""

        return {
            "wake_light": {
                item["name"].lower(): idx
                for idx, item in enumerate(self._get("files/lightthemes").values())
                if item["name"]
            },
            "dusk_light": {
                item["name"].lower(): idx
                for idx, item in enumerate(self._get("files/dusklightthemes").values())
            },
            "wake_sound": {
                item["name"].lower(): idx + 1
                for idx, item in enumerate(self._get("files/wakeup").values())
                if item["name"]
            },
            "dusk_sound": {
                item["name"].lower(): idx + 1
                for idx, item in enumerate(self._get("files/winddowndusk").values())
                if item["name"]
            },
        }

    def get_sensor_data(self):
        """Get sensor data as a dictionary."""
        data = self._get("wusrd")
        return {
            "temperature": data.get("mstmp"),
            "humidity": data.get("msrhu"),
            "luminance": data.get("mslux"),
            "noise": data.get("mssnd"),
        }

    def get_alarm_status(self):
        """Get alarm status"""
        return self._get("wusts")

    def get_light_data(self):
        """Get light data"""
        return self._get("wulgt")

    def get_sunset_data(self):
        """Get sunset data"""
        return self._get("wudsk")

    def get_enabled_alarms(self):
        """Get enabled alarms"""
        return self._get("wualm/aenvs")

    def get_time_alarms(self):
        """Get time alarms"""
        return self._get("wualm/aalms")

    def get_snooze_time(self):
        """Get snooze time"""
        return self._get("wualm")

    def get_player_status(self):
        """Get player status"""
        return self._get("wuply")

    def modify_light(self, payload: dict) -> dict:
        """Set light data"""
        if (
            "wucrv" in payload
        ):  # Some Wake-ups lights don't work with wucrv, remove key if exists
            payload.pop("wucrv")
        return self.put("wulgt", payload=payload)

    def modify_sunset(self, payload: dict) -> dict:
        """Set sunset data"""
        return self.put("wudsk", payload=payload)

    def modify_player(self, payload: dict) -> dict:
        """Set player control data"""
        return self.put("wuply", payload=payload)

    def modify_alarm_details(self, payload: dict) -> dict:
        """Set alarm control data"""
        return self.put("wualm", payload=payload)

    def modify_running_alarm(self, payload: dict) -> dict:
        """Set alarm control data"""
        return self.put("wualm/alctr", payload=payload)

    def modify_alarm_wake_up_configuration(self, payload: dict) -> dict:
        """Set alarm wake up data"""
        return self.put("wualm/prfwu", payload=payload)

    def modify_alarm_status(self, payload: dict) -> dict:
        """Set alarm wake up status"""
        return self.put("wusts", payload=payload)
